from .. import auth

token = "eyJhbGciOiJFZERTQSIsImtpZCI6InNpZy0xNjQ0OTM3MzYwIn0.eyJqdGkiOiI1YWJhOGJhNi05MGQ1LTQ4N2QtOWEyNy1iYzA0N2U1NTdjYzAiLCJhdWQiOlsiYXBwOjI4NDYzNjA3NjI0MjM3MTA4NSIsImh0dHBzOi8vYXBpLnJvd25kLmlvIl0sInN1YiI6Imdvb2dsZS1vYXV0aDJ8MTE0ODk1MTIyNzk1NDUyMTI3Mjc2IiwiaWF0IjoxNjQ5MzA0NjY5LCJodHRwczovL2F1dGgucm93bmQuaW8vYXBwX3VzZXJfaWQiOiIxN2E4MjE3Zi04MmUzLTQwZGItODU1Ni0xNmIxMGQzNThkYTgiLCJodHRwczovL2F1dGgucm93bmQuaW8vaXNfdmVyaWZpZWRfdXNlciI6dHJ1ZSwiaXNzIjoiaHR0cHM6Ly9hcGkucm93bmQuaW8iLCJleHAiOjE2NDkzMDgyNjl9.qct6lD7AJW97gjYCYyfvyArbVlE8jf76joCHcEbEgNkE1d-EPSgBIvA9MoRSP2Ezm4-uymCHBC7cenlObXBZBw"
auth.client.validate_jwt(token)


"""
Settings for Rownd are all namespaced in the ROWND setting.
For example your project's `settings.py` file might look like this:
ROWND = {
    'APP_KEY': '<your app key>',
    'APP_SECRET': '<your app secret>',
}
This module provides the `rownd_setting` object, that is used to access
Rownd settings, checking for user settings first, then falling
back to the defaults.
"""
import os
from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

DEFAULTS = {
    'API_URL': 'https://api.rownd.io',
    'APP_KEY': os.environ.get("ROWND_APP_KEY") or None,
    'APP_SECRET': os.environ.get("ROWND_APP_KEY") or None,
}


# List of settings that may be in string import notation.
IMPORT_STRINGS = []


# List of settings that have been removed
REMOVED_SETTINGS = []


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class RowndSettings:
    """
    A settings object that allows Rownd settings to be accessed as
    properties. For example:
        from rownd.settings import rownd_settings
        print(rownd_settings.API_URL)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    Note:
    This is an internal class that is only compatible with settings namespaced
    under the ROWND name. It is not intended to be used by 3rd-party
    apps, and test helpers like `override_settings` may not work as expected.
    """
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'ROWND', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def __check_user_settings(self, user_settings):
        SETTINGS_DOC = "https://docs.rownd.io/rownd/sdk-reference/web/django"
        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError("The '%s' setting has been removed. Please refer to '%s' for available settings." % (setting, SETTINGS_DOC))
        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


rownd_settings = RowndSettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_rownd_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'ROWND':
        rownd_settings.reload()


setting_changed.connect(reload_rownd_settings)


from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import User

from rest_framework import authentication, exceptions

from rownd_django.auth import client

class RowndAuthError(Exception):
    pass

class RowndAuthBackend:
    def _authenticate(self, token: str):
        if not token:
            return None
            
        try:
            token_info = client.validate_jwt(token)
            user = User.objects.get(username=token_info["sub"])
            return user
        except User.DoesNotExist:
            # fetch user from Rownd
            try:
                rownd_user = client.fetch_user(token_info["https://auth.rownd.io/app_user_id"])
                user = User.objects.get(email=rownd_user["data"]["email"])
                return user
            except User.DoesNotExist:
                # create the user
                try:
                    user = User(username=token_info["sub"], email=rownd_user["data"]["email"], first_name=rownd_user["data"].get("first_name", ""), last_name=rownd_user["data"].get("last_name", ""))
                    user.save()
                    return user
                
                except:
                    raise RowndAuthError("Failed to locate or create user")
        except Exception as e:
            # The token didn't pass validation or
            # we couldn't find the Rownd user or
            # something went wrong creating the local user
            raise RowndAuthError(e)

class RowndAuthenticationBackend(RowndAuthBackend, BaseBackend):
    def authenticate(self, request, token=None):
        try:
            return self._authenticate(token)
        
        except Exception as e:
            print(e)
            return None

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

class RowndApiAuthentication(RowndAuthBackend, authentication.BaseAuthentication):
    def authenticate(self, request):
        try:
            token = request.headers.get("authorization")
            token = token.split(" ")[1]
            user = self._authenticate(token)
            return (user, None)

        except RowndAuthError as e:
            raise exceptions.AuthenticationFailed(e)

        except Exception as e:
            raise exceptions.AuthenticationFailed("Invalid or malformed token")


import os
import re
import shutil
import sys
from setuptools import find_packages, setup

def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)

version = get_version('rownd_django')

if sys.argv[-1] == 'publish':
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    if os.system("twine check dist/*"):
        print("twine check failed. Packages might be outdated.")
        print("Try using `pip install -U twine wheel`.\nExiting.")
        sys.exit()
    os.system("twine upload dist/*")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    shutil.rmtree('dist')
    shutil.rmtree('build')
    shutil.rmtree('rownd_django.egg-info')
    sys.exit()

setup(
    version=version,
    packages=find_packages(exclude=['tests*']),
)


import json
from django.http import HttpResponse
from django.contrib.auth import authenticate, login
from django.views.decorators.http import require_POST

@require_POST
def session_authenticate(request):
    if request.user.is_authenticated:
        return HttpResponse(status=200, content=json.dumps({ 'message': 'Already authenticated' }))

    token = request.headers.get("Authorization")

    if not token:
        return HttpResponse(status=401)

    token = token.split(" ")[1]
    
    user = authenticate(request, token=token)
    if user is not None:
        login(request, user)
        return HttpResponse(status=200, content=json.dumps({ 'message': 'Authentication successful', 'should_refresh_page': True }))
    else:
        return HttpResponse(status=401)


import jwt
import requests
from rownd_django.auth import jwks_client
from rownd_django.settings import rownd_settings

__app_config = None

def fetch_app_config():
    global __app_config
    r = requests.get(f"{rownd_settings.API_URL}/hub/app-config", headers={
        'x-rownd-app-key': rownd_settings.APP_KEY,
    })

    if r.status_code != 200:
        raise Exception(f"Failed to fetch app config: {r.json()}")

    __app_config = r.json()
    return __app_config

def validate_jwt(token: str, audience: list[str]=[]):
    global __app_config
    if __app_config is None:
        fetch_app_config()

    jwks_client = fetch_jwks()
    signing_key = jwks_client.get_signing_key_from_jwt(token)

    default_audience = f'app:{__app_config["app"]["id"]}'
    audience.append(default_audience)

    try:
        return jwt.decode(
            token,
            signing_key.key,
            algorithms=["EdDSA"],
            audience=audience
        )
    except jwt.exceptions.InvalidAudienceError:
        raise jwt.exceptions.InvalidAudienceError(f"Invalid audience. Expected {audience}")

    except Exception as e:
        raise e

def fetch_jwks() -> jwks_client.PyJWKClient:
    url = f"{rownd_settings.API_URL}/hub/auth/keys"
    return jwks_client.PyJWKClient(url)

def fetch_user(user_id: str):
    global __app_config
    r = requests.get(f'{rownd_settings.API_URL}/applications/{__app_config["app"]["id"]}/users/{user_id}/data', headers={
        "x-rownd-app-key": rownd_settings.APP_KEY,
        "x-rownd-app-secret": rownd_settings.APP_SECRET,
    })

    if r.status_code != 200:
        raise Exception(f"Failed to fetch user: {r.json()}")

    return r.json()

if __name__ == "__main__":
    rownd_settings.APP_KEY = "<key>"
    rownd_settings.APP_SECRET = "<secret>"
    token = "<token>"
    token_info = validate_jwt(token)
    print(token_info)
    user_info = fetch_user(token_info["https://auth.rownd.io/app_user_id"])
    print(user_info)
    print("Success!")
